function [ab, clErr, itrErrors] = learn(ab, trnExamples, targetOuts, ...
    nStages, verbose, crossvalset, crossvalout, reqDetRate, reqErrBnd)
% function [ab, err] = learn(ab, trnExamples, targetOuts, nStages, reqDetRate)
%   learning function for the adaBooster
%
%   Inputs:
%       trnExamples: examples for training the classifier.
%       targetOuts: target classification outputs. its size must be the
%           same as the number of examples.
%       nStages: number of boosting stages required [default 10]
%       reqDetRate: target detection rate for this classifier
%           which is the percentage of positive examples that are classified
%       reqErrBnd: bound on miss classification error, training stops when 
%           it is reached or the max number of stages is reached.
%   Outputs:
%       ab: trained adaBooster
%       clErr: classification error of the trained classifier

%% ---
% fprintf like function that works only if condition(first parameter) is true
function printfTrue(condition, varargin)
    if condition
        fprintf(varargin{:});
    end
end

%% Check Validity of Input
if isempty(ab.weakCl)
    error('A weak classifier has not been specified');
end

if nargin < 3
    error('incorrect number of arguments');
end

if nargin < 4 || isempty(nStages),
    nStages = 10;
end

if nargin < 5
    verbose = false;
end

if nargin > 5 && nargin < 7
    error('cross validation output is not given');
end

useCrossValidationSet = false;

if nargin >= 6 && ~isempty(crossvalset) && ~isempty(crossvalout)
    useCrossValidationSet = true;
end

if nargin < 8 || isempty(reqDetRate)
    reqDetRate = NaN;
end

if nargin >= 9 && ~isempty(reqErrBnd)
    ab.errBound = reqErrBnd;
end

if nargout > 2 
    if isinf(nStages)
        errors('Cant record errors with infinite number of stages');
    else
        itrErrors = zeros(nStages, 1);
    end
end

printfTrue(verbose, 'AdaBoost\n');
printfTrue(verbose, '========\n');

%% Initialize Example Weights
if isempty(ab.lastExWeights)
    % that classifier was not trained before
    
    % initialize weights to be uniform over each class
    wts = zeros(length(targetOuts), 1);
    nc = getNumClasses(ab);
    for c = 1:nc
        inds = targetOuts == c;
        numc = sum(inds);
        wts(inds) = 1 / (nc * numc);
    end
%     % initialize weights to be uniform over all training examples
%     wts = repmat(1 / length(targetOuts), length(targetOuts), 1);
else
    % classifier was trained before for a different number of stages
    % initialize weights to the weights of the last stage trained before
    wts = ab.lastExWeights;
end;

%% Boosting Loop - Begin
%%
clErr = Inf;
ab.thresh = 0;
while (isinf(nStages) && clErr > ab.errBound) || ...
      (~isinf(nStages) && length(ab.clsWeights) < nStages),

    printfTrue(verbose, '\n\t\t\t==================\n');
    printfTrue(verbose, '\t\t\tBoosting Stage # %d\n', length(ab.clsWeights) + 1);
    printfTrue(verbose, '\t\t\t==================\n\n');

    if useCrossValidationSet
        % the weak classifier must support cross validation set
        [trndCl, err] = learn(ab.weakCl, trnExamples, targetOuts, wts, ...
            crossvalset, crossvalout);
    else
        [trndCl, err] = learn(ab.weakCl, trnExamples, targetOuts, wts);
    end
    
    if err < 1e-9, % that is to avoid division by zero later on
        % Division by zero may occur also if err is almost 1.
        % However, I assume here that the weak classifier does not produce
        %   error rate higher than 1-(1/number of classes).
        err = 1e-9;
    end

    printfTrue(verbose, ... 
            '\t\t\tWeak Classifier has been trained, err = %f\n', err);

    beta = err / (1 - err);
    % add the trained classifier to the set of trained classifiers
    ab.trndCls = {ab.trndCls{:}, trndCl}; %#ok<CCAT>
    % compute the weight of this classifier
    alpha = log(1 / beta) + log(getNumClasses(ab) - 1);
    ab.clsWeights = [ab.clsWeights, alpha];
    %ab.thresh = sum(ab.clsWeights) * (posVal + negVal) / 2;
    % update the weights of training examples to be ready for the next
    % stage

    outs = computeOutputs(trndCl, trnExamples);

    e = outs ~= targetOuts;
    wts = wts .* exp(alpha .* e);
    % normalize the weights
    wts = wts / sum(wts);
    if isinf(nStages) || nargout > 2
        % compute boosted classification error only if the number of stages
        % is not specified, otherwise, no need to compute it now
        abOuts = computeOutputs(ab, trnExamples);
        clErr = sum(abOuts ~= targetOuts) / length(abOuts);
        printfTrue(verbose, ...
            '\t\t\tCurrently, boosted classifier''s error = %f\n', clErr);
        if ~isinf(nStages)
            itrErrors(length(ab.clsWeights) ,1) = clErr;
        end
    end
end
if ~isinf(nStages) && nargout < 3
    % compute boosted classification error only if the number of stages
    % is specified, otherwise, it is already compute in the boosting loop
    abOuts = computeOutputs(ab, trnExamples);
    clErr = sum(abOuts ~= targetOuts) / length(abOuts);
end
%% Boosting Loop - End
%% Set Number of Stages and Last Set of Weighs Computed
ab.lastExWeights = wts;
%% Adjust Threshold if Necessary (valid only in binary classification)
if getNumClasses(ab) == 2
    if ~isnan(reqDetRate),
        [ab, clErr] = adjustThreshold(ab, trnExamples, targetOuts, ...
                                          reqDetRate);
    else
        ab.detectionRate = ...
            sum((targetOuts == getPosVal(ab)) & (abOuts == getPosVal(ab))) / ...
            sum((targetOuts == getPosVal(ab)));
    end
end

printfTrue(verbose, ...
        '\t\t\tAdaBoost Training is Done with err = %f\n', clErr);
end
