function [lb, clErr] = learn(lb, trnExamples, targetOuts, nStages, reqDetRate, reqErrBnd)
% function [lb, err] = learn(lb, trnExamples, targetOuts, nStages, reqDetRate)
%   learning function for the Logit boost 
%
%   Inputs:
%       trnExamples: examples for training the classifier.
%       targetOuts: target classification outputs. its size must be the
%           same as the number of examples.
%       nStages: number of boosting stages required [default 10]
%       reqDetRate: target detection rate for this classifier
%           which is the percentage of positive examples that are classified
%       reqErrBnd: bound on miss classification error, training stops when 
%           it is reached or the max number of stages is reached.
%   Outputs:
%       ab: trained logitBoost
%       clErr: classification error of the trained classifier

%% ----
disp('============');
disp('Logit Boost')
disp('============');
%% Check Validity of Input
if nargin < 3
    error('incorrect number of arguments');
end

if nargin < 4 || isempty(nStages),
    nStages = 10;
end

if nargin < 5 || isempty(reqDetRate)
    reqDetRate = NaN;
end

if nargin >= 6 && ~isempty(reqErrBnd)
    lb.errBound = reqErrBnd;
end

%% Initialize Example Weights
if isempty(lb.lastExWeights)
    wts = ones(length(targetOuts), 1) * 1/length(targetOuts);
else
    % classifier was trained before for a different number of stages
    % initialize weights to the weights of the last stage trained before
    disp('classifier was trained before');
    wts = lb.lastExWeights;
end;

%% Boosting Loop - Begin
%%
clErr = Inf;
lb.thresh = 0;

%handel cell input
trainData = trnExamples;
if iscell(trainData),
    trainData = cell2mat(trainData);
end

% init F and P
J = getNumClasses(lb);
if lb.nStages == 0
    lb.F = zeros(J, size(trainData,2) + 1);      % (J,n)
    lb.P = ones(size(trainData,1), J) * 1/J;        % (m,n)
end

iStage = lb.nStages;   %stage iterator

while (isinf(nStages) && clErr > lb.errBound) || ...
      (~isinf(nStages) && iStage < nStages),

    fprintf('Iteration %d\n', iStage);
    z = zeros(size(trainData,1), J);
    w = zeros(size(trainData,1), J);
    Fj = zeros(J, size(trainData,2) + 1);
    sumFj = zeros(J, size(trainData,2) + 1);
    breakLoop = false;
    for j = 1:J
        JVec = ones(size(trainData,1),1) * j;
        YStar = JVec == targetOuts;
        
        z(:,j) = (YStar - lb.P(:,j)) ./ ( (lb.P(:,j)) .* ( 1 - lb.P(:,j)) );
        w(:,j) = ( (lb.P(:,j)) .* ( 1 - lb.P(:,j)) );
        
        % break if nan values was detected
        if (sum(isnan(w(:,j))) > 0 || sum(isnan(z(:,j))) > 0)
            breakLoop = true;
            break;
        end
        
        regressor = learn(lb.regressor, trainData, z(:,j), w(:,j));
        Fj(j,:) = getWeights(regressor);
        
        % break if nan values was detected
        if (sum(isnan(Fj(j,:))) > 0)
            breakLoop = true;
            break;
        end
        
        sumFj(j,:) = sumFj(j,:) + Fj(j,:);
    end
  
    if (~breakLoop)
        Fj = ((J-1)/J)*(Fj - (1/J) * sumFj );
        lb.F = lb.F + Fj;
    else
        fprintf('next iterations (after %d) will have no effects due to nan\n' ...
         ,iStage+1);

    end
    
    
    Fvalues = [trainData ones(size(trainData,1),1)] * (lb.F)';
    
    eFvalues = exp(Fvalues);
    eps = 1e-6;
    lb.P = bsxfun(@times, eFvalues, 1 ./ max(sum(eFvalues,2), eps ));
    
    %-------------------------------------------------
    if isinf(nStages)
        % compute boosted classification error only if the number of stages
        % is not specified, otherwise, no need to compute it now
        
        abOuts = computeOutputs(lb, trainData);

        clErr = sum(abOuts ~= targetOuts) / length(abOuts);
        fprintf('\t\t\tCurrently, boosted classifier''s error = %f\n', clErr);
    end
    
    iStage = iStage + 1;
    lb.nStages = lb.nStages + 1;
end
if ~isinf(nStages)
    % compute boosted classification error only if the number of stages
    % is specified, otherwise, it is already compute in the boosting loop    
    abOuts = computeOutputs(lb, trainData);
    clErr = sum(abOuts ~= targetOuts) / length(abOuts);
end
%% Boosting Loop - End
%% Set Number of Stages and Last Set of Weighs Computed
lb.lastExWeights = wts;

%% Adjust Threshold if Necessary (valid only in binary classification)
if getNumClasses(lb) == 2
    if ~isnan(reqDetRate),
        if useGlobalExamples,
            [lb, clErr] = adjustThreshold(lb, [], targetOuts, reqDetRate);
        else
            [lb, clErr] = adjustThreshold(lb, trainingExamples, targetOuts, ...
                                          reqDetRate);
        end
    else
        lb.detectionRate = ...
            sum((targetOuts == getPosVal(lb)) & (abOuts == getPosVal(lb))) / ...
            sum((targetOuts == getPosVal(lb)));
    end
end
fprintf('\t\t\tLogit Boost Training is Done with err = %f\n', clErr);
