function [outs, realOuts] = computeOutputs(rab, examples, extraArg)
% function [outs] = computeOutputs(ab, examples)
%   computes the classification outputs of realAdaBoost for the given examples
%
%   Inputs:
%       rab : real ada boost classifier instance
%       examples: set of example patterns to be classified
%       extraArg: an extra argument to be used if needed
%   
%   Returns:
%       outs [number of instances X 1] :
%            output class labels for given example instances
%       realOuts [number of Instances X number of classes K] : 
%             h(x) hypothesis value for each class (for each instance)

if isempty(rab.weakCl)
    error('A weak classifier is not specified');
end
if isnan(rab.thresh) && ~rab.perfect
    error('RealAdaBoost classifier has not been trained');
end

if nargin == 1
    error('incorrect number of arguments');
elseif nargin == 2
    extraArgs = {examples};
else
    if iscell(extraArg),
        extraArgs = {examples, extraArg{:}}; %#ok<CCAT>
    else
        extraArgs = {examples, extraArg};
    end
end

% -----------
K = getNumClasses(rab.trndCls{1});

% discrete ada boost
% realOuts(sub2ind(sz, rowInds, wkClOuts)) = ab.clsWeights(1);

% real ada boost
function h = calcH(prob)
    % SAMME.R step 2c, calculation of h
    log_prob = log(prob);
    sum_log_prob = sum(log_prob,2);
    h = (K-1) * (log_prob - (((1/K) * sum_log_prob) * ones(1,K)));
end

if iscell(examples)
    realOuts = zeros(length(examples), K);
else
    realOuts = zeros(size(examples, 1), K);
end

for i = 1:length(rab.trndCls),
    [~, prob] = computeOutputs(rab.trndCls{i}, extraArgs{:});
    
    % discrete ada boost
    % realOuts(inds) = realOuts(inds) + ab.clsWeights(i);
    
    %real ada boost    
    realOuts = realOuts + calcH(prob);
end

[~, outs] = max(realOuts, [], 2);

end