function [rab, clErr, itrErrors] = learn(rab, trnExamples, targetOuts, ...
    nStages, verbose, crossvalset, crossvalout, reqDetRate, reqErrBnd)
% function [ab, err, itrErrors] = learn(ab, trnExamples, targetOuts, ...
%    nStages, crossvalset, crossvalout, reqDetRate, reqErrBnd)
%
%   learning function for the realAdaBooster
%
%   Inputs:
%       1 rab : real ada boost classifier instance
%       2 trnExamples: examples for training the classifier.
%       3 targetOuts: target classification outputs. its size must be the
%           same as the number of examples.
%       4 nStages: number of boosting stages required [default 10]
%       5 verbose: prints into while classifier is working
%       6 crossvalset : cross validation set
%       7 crossvalout : output of cross validation set
%       8 reqDetRate: target detection rate for this classifier
%           which is the percentage of positive examples that are classified
%       9 reqErrBnd: bound on miss classification error, training stops when 
%           it is reached or the max number of stages is reached.
%   Outputs:
%       ab: trained realAdaBooster
%       clErr: classification error of the trained classifier
%       itrErrors : total error of real adaboost classifier at each iteration

%% ----

% fprintf like function that works only if condition(first parameter) is true
function printfTrue(condition, varargin)
    if condition
        fprintf(varargin{:});
    end
end

%% Check Validity of Input
if isempty(rab.weakCl)
    error('A weak classifier has not been specified');
end

if nargin < 3
    error('incorrect number of arguments');
end

if nargin < 4 || isempty(nStages),
    nStages = 10;
    rab.nStages = 10;
else
    rab.nStages = nStages;    
end

if nargin < 5
    verbose = false;
end

if nargin > 5 && nargin < 7
    error('cross validation output is not given');
end

useCrossValidationSet = false;

if nargin >= 6 && ~isempty(crossvalset) && ~isempty(crossvalout)
    useCrossValidationSet = true;
end

if nargin < 8 || isempty(reqDetRate)
    reqDetRate = NaN;
end

if nargin >= 9 && ~isempty(reqErrBnd)
    rab.errBound = reqErrBnd;
end

if nargout > 2 
    if isinf(nStages)
        errors('Cant record errors with infinite number of stages');
    else
        itrErrors = zeros(nStages, 1);
    end
end

printfTrue(verbose, 'RealAdaBoost\n');
printfTrue(verbose, '============\n');

%% Initialize Example Weights
if isempty(rab.lastExWeights)
    printfTrue(verbose, 'RealAda boost Classifier wasnt trained before\n' ,1);
    wts = ones(length(targetOuts), 1) * 1/length(targetOuts);
else
    % classifier was trained before for a different number of stages
    % initialize weights to the weights of the last stage trained before
    printfTrue(verbose, 'Real AdaBoost has been trained before\n');
    wts = rab.lastExWeights;
end;

%% Boosting Loop - Begin

clErr = Inf;
rab.thresh = 0;

iStage = length(rab.trndCls) + 1;

while (isinf(nStages) && clErr > rab.errBound) || ...
      (~isinf(nStages) && iStage <= nStages),
  
    printfTrue(verbose, '\n\t\t\t==========================\n');
    printfTrue(verbose, '\t\t\tRealAda Boosting Stage # %d\n', iStage);
    printfTrue(verbose, '\t\t\t==========================\n');
    
    % First : Fit classifier with prob estimates
    if useCrossValidationSet
        % the weak classifier must support cross validation set
        [trndCl, err] = learn(rab.weakCl, trnExamples, targetOuts, wts, ...
            crossvalset, crossvalout);
    else
        [trndCl, err] = learn(rab.weakCl, trnExamples, targetOuts, wts);
    end
    
    % avoid division by zero
    if err < 1e-9, 
        err = 1e-9;
    end
    
    printfTrue(verbose, ...
            '\t\t\tWeak Classifier has been trained, err = %f\n', err);
    
    % add the trained classifier to the set of trained classifiers
    rab.trndCls = {rab.trndCls{:}, trndCl};
    % calculate classifier weight
    beta = err / (1 - err);
    alpha = log(1 / beta) + log(getNumClasses(rab) - 1);
    rab.clsWeights = [rab.clsWeights, alpha];
    
    % Second : Get probabilty estimates of new weak classifier
    [~, prob] = computeOutputs(trndCl, trnExamples);
    
    % Third: update the weights of training examples to be ready for the next
    % stage
    
    % construct the Y Coding matrix (SAMME equation number 7)
    K = getNumClasses(rab);
    Y_CODE = ones(length(targetOuts), K) * (-1/(K-1));
    Y_CODE(sub2ind(size(Y_CODE), (1:length(targetOuts))' , targetOuts)) = 1;
        
    % avoid zero prob for the log and also negative prob (due to negative wts)
    prob(prob <= 0) = 1e-5;
    
    % dot product of Y_CODE and log Prob matrix multiplied with -1 * (k-1)/k
    estimator_wts = -1 * ((K-1)/K) * sum(Y_CODE .* log(prob), 2);

    %wts = wts .* exp(estimator_wts);
    wts = wts .* exp( (estimator_wts .* (+(estimator_wts < 0) |+(wts > 0) ) ) );
    
    % finally normalize wts
    wts = wts ./ sum(wts);
    
    if isinf(nStages) || nargout > 2
        % compute boosted classification error only if the number of stages
        % is not specified, otherwise, no need to compute it now
        abOuts = computeOutputs(rab, trnExamples);
        clErr = sum(abOuts ~= targetOuts) / length(abOuts);
        printfTrue(verbose, '\t\t\tCurrently, boosted classifier''s error = %f\n', clErr);
        if ~isinf(nStages)
            itrErrors(iStage ,1) = clErr;
        end
    end
    
    iStage = iStage + 1;
end

if ~isinf(nStages) && nargout < 3
    % compute boosted classification error only if the number of stages
    % is specified, otherwise, it is already compute in the boosting loop

    abOuts = computeOutputs(rab, trnExamples);
    clErr = sum(abOuts ~= targetOuts) / length(abOuts);
end
%% Boosting Loop - End
%% Set Number of Stages and Last Set of Weighs Computed
rab.lastExWeights = wts;

printfTrue(verbose, '\t\t\tRealAdaBoost Training is Done with err = %f\n', clErr)
end
