function [svmCl, clErr] = learn(svmCl, trnExamples, targetOuts, wts)
% LEARN learning function of svm classifier
% function [svmCl, err] = learn(svmCl, trnExamples, targetOuts)
%   learning function for the SVM classifier
%
%   Inputs:
%       trainingExamples: examples for training the classifier.
%       targetOuts: target classification outputs. its size must be the
%           same as the number of examples.
%       wts: weights of training examples (used to compute the weighted
%           classification error)
%   Outputs:
%       svmCl: trained SVMClassifier
%       clErr: classification error of the trained classifier

%% Deal with Cell Array Input
if iscell(trnExamples),
    trnExamples = cell2mat(trnExamples);
end

%% Create and Train the SVM
% construct the weight vector for the svm
nExamples = numel(targetOuts);
if nargin < 4 || isempty(wts),
    wts = zeros(length(targetOuts), 1);
    nc = getNumClasses(svmCl);
    for c = 1:nc
        inds = targetOuts == c;
        numc = sum(inds);
        wts(inds) = 1 / (nc * numc);
    end
%     wts = ones(nExamples, 1) / nExamples;
end
mxWts = max(wts);
mnWts = min(wts);
if mnWts == mxWts
    % make them all 10
    svmWts = repmat(10, nExamples, 1);
else
    % use linear mapping from 1 to 20
    svmWts = 19 * (wts - min(wts)) / (max(wts) - min(wts)) + 1;
end
% svmWts = max(20 * wts / max(wts), 1e-2);
% NOTE: libsvm-weights does not work properly if weights are small
% values. In the example they have, all the weights are between 1 and
% 20. It seems that when the weights add up to one, the computed
% misclassification error becomes too small and the trainer terminates
% without learning anything. So, here, I adjust the weights to be 20 at
% maximum. Setting the minimum weight to 1e-2 is set to match a
% restriction on a different implementation for SVM.
% NOTE 2: I changed the weighting so that the minimum is 1, not 1e-2,
% because the behavior of SVM was very poor when the difference between the
% maximum and minimum weight is very large.

% create and train the SVM

svmCl.trainedSVM = svmtrain(svmWts, targetOuts, trnExamples, svmCl.libSvmTrnOpts);

%% Compute Classification Error
if nargout > 1
    [outs, ~, ~] = svmpredict(targetOuts, trnExamples, svmCl.trainedSVM, svmCl.libSvmPrdOpts);
    if nargin < 4,
        clErr = sum(outs ~= targetOuts) / nExamples;
    else
        clErr = (outs ~= targetOuts)' * wts;
    end
    if isnan(clErr),
        pause;
    end
end
